@@ -3,6 +3,8 @@ from __future__ import annotations
 import asyncio
 import json
 import time
+from copy import deepcopy
+from logging import Logger, getLogger
 from os.path import dirname, isfile, realpath
 from pathlib import Path
 from typing import Any, Callable, List, Tuple, Type, Union, cast
@@ -16,9 +18,11 @@ from eth_typing.encoding import HexStr
 from hexbytes import HexBytes
 from typing_extensions import Self
 from web3 import Web3
+from web3._utils.method_formatters import BLOCK_FORMATTERS
 from web3.contract.contract import Contract, ContractFunction, ContractFunctions
 from web3.exceptions import TransactionNotFound
 from web3.gas_strategies import rpc
+from web3.middleware.validation import BLOCK_VALIDATORS
 from web3.types import (
     BlockData,
     BlockIdentifier,
@@ -32,7 +36,11 @@ from web3.types import (
 )
 from websockets.client import connect
 
-from web3client.exceptions import TransactionTooExpensive, Web3ClientException
+from web3client.exceptions import (
+    ProviderNotSet,
+    TransactionTooExpensive,
+    Web3ClientException,
+)
 from web3client.helpers.subscribe import parse_notification, subscribe_to_notification
 from web3client.middlewares.rpc_log_middleware import (
     BaseRpcLog,
@@ -47,36 +55,72 @@ from web3client.types import (
 
 class BaseClient:
     """
-    Client to interact with a blockchain, with smart contract
-    support.
+    Interact with a blockchain, with smart contract support.
 
-    The client is a wrapper intended to make the Web3 library
-    easier to use.
-    """
+    Smart contract support
+    ----------------------
+
+    If you specify a contract address and an ABI, you can then use the
+    `self.call(<function>)` and `self.transact(<function>)` methods to interact
+    with the smart contract.  The functions of the smart contract are available
+    as attributes of `self.functions`.
+
+    The easiest way to use web3client with smart contracts, is to subclass
+    BaseClient.  See for example the `Erc20Client` class, to interact with
+    ERC-20 tokens, and `compound_v2_client.py`, which contains classes to
+    interact with Compound V2 pools and comptroller.
+
+
+    Transaction types & gas
+    -----------------------
+
+    The client supports both EIP-1559 and legacy transactions.  EIP-1559 transactions
+    are the default.  Whenever possible, gas will be estimated.  The rationale is to
+    spare the user the need to specify gas parameters.
+
+    The transaction type is set with the `tx_type` attribute, which can be set either
+    at the class level or instance level.  Its possible values are:
+
+    - tx_type = None (default): The client will try to infer the transaction type from
+        the node.  If the node does not support EIP-1559, it will fall back to legacy
+        transactions.
+    - tx_type = 2: Will send EIP-1559 transactions.
+        The miner's tip (maxMaxFeePerGas) can be set at the class level, instance
+        level or function level.  If not set, it will be set to 0.01 Gwei by default.
+        The max fee (maxMaxFeePerGas) is estimated according to the usual formula
+        maxMaxFeePerGas = 2 * baseFee + maxPriorityFeePerGas, where baseFee is the
+        base fee of the last block.
+    - tx_type = 0: Will send legacy transactions.  Gas price, if not provided,
+        will be estimated using eth_gasPrice.
+    - tx_type = 1: This value indicates a legacy transaction with support for
+        access lists (EIP2930 transactions).  The client does not support access lists
+        yet, therefore we will treat this as a legacy transaction.
+
+    The gas limit is set indipendently than the transaction type.  If not set, it will
+    be estimated using eth_gasEstimate.  If set, it will be used as is, and no estimation
+    will be performed.
+
+    More details on transaction types & gas on Infura docs:
+    https://docs.infura.io/networks/ethereum/concepts/transaction-types
+
+
+    Set a max cost for gas
+    ----------------------
+
+    You can set a maximum cost for gas, to avoid paying too much for a transaction.
+    To do so, set the `upper_limit_for_base_fee_in_gwei`. If the base fee from the
+    last block is larger than this limit, the client will raise a `TransactionTooExpensive`
+    exception.  By default, there is no limit.
+
+    The `upper_limit_for_base_fee_in_gwei` parameter works for legacy transactions,
+    too.  In this case, the gas price, instead of the base fee.
+
+
+    RPC logging
+    -----------
+
+    The client supports flexible logging of RPC calls, via the `rpc_logs` argument.
 
-    # Attributes that can be set in the constructor or
-    # overridden by subclasses
-    node_uri: str = None
-    """RPC node to use.  Set it to None for a uninitialized client."""
-    chain_id: int = None
-    """ID of the chain.  If not given, it will be inferred from the node"""
-    tx_type: int = None
-    """Type of transaction. type=2 means an EIP-1599 transaction. More details here > https://docs.infura.io/infura/networks/ethereum/concepts/transaction-types"""
-    private_key: str = None
-    """Private key to use to send transactions"""
-    max_priority_fee_in_gwei: float = None
-    """"Miner's tip, relevant only for type-2 transactions (default is 1)"""
-    upper_limit_for_base_fee_in_gwei: float = None
-    """Raise an exception if baseFee is larger than this (default is no limit)"""
-    abi: dict[str, Any] = None
-    """ABI of smart contract; to read from a JSON file, use class method get_abi_json()"""
-    contract_address: str = None
-    """Address of smart contract"""
-    middlewares: List[Middleware] = None
-    """Ordered list of web3.py middlewares to use"""
-    rpc_logs: List[BaseRpcLog] = None
-    """Where to log RPC calls.
-    
     To log all RPC call, set `rpc_logs=[PythonLog()]`.  This will make use of
     the logger named 'web3client.RpcLog', with logLevel=INFO.  To log to a different
     logger, specify it: `rpc_logs=[PythonLog(logger=logging.getLogger("my_logger"))]`.
@@ -94,9 +138,64 @@ class BaseClient:
         client.send_eth("0xde0b295669a9fd93d5f28d9ec85e40f4cb697bae", 1)
         print(rpc.log.entries) # will contain 2 entries: request & response
         ```
-    
+
     One can also log to file, or to a database.  For more details, see the docs of the
-    rpc_log_middleware module."""
+    rpc_log_middleware module.
+
+    Please note that subclasses can define a class attribute `rpc_logs` to set a default
+    value for `rpc_logs` for all instances of the subclass.
+
+    POA support
+    -----------
+
+    web3client supports PoA chains like BnB, Polygon POS, Avalanche and Scroll
+    out of the box.  We do so by disabling the checks that web3.py does on the
+    `extraData` block property.
+
+    If for any reason you want to keep the `extraData` checks, set `add_poa_support=False`.
+    Please note that if you keep the checks, and want to use a PoA chain, then
+    you will also need to set `middlewares=[web3.middleware.geth_poa_middleware]`.
+    Using the middleware instead of `add_poa_support` would also rename the
+    `extraData` field to `proofOfAuthorityData`.
+
+    More details:
+        - http://web3py.readthedocs.io/en/stable/middleware.html#proof-of-authority
+        - https://openethereum.github.io/Proof-of-Authority-Chains
+    """
+
+    # Attributes that can be either set at instantiation or at the class level,
+    # with their default values.  Being able to set these attributes at the class
+    # level makes it possible to change them by subclasses, or dynamically, using
+    # monkey patching.
+    node_uri: str = None
+    """RPC node to use.  Set it to None for a uninitialized client."""
+    chain_id: int = None
+    """ID of the chain.  If not given, it will be inferred from the node"""
+    tx_type: int = None
+    """Type of transaction.  Leave empty to infer it from the node.  More details in class docstring."""
+    max_priority_fee_in_gwei: float = 0.01
+    """"Miner's tip, in gwei.  Relevant only for type-2 transactions.  Default is 0.01 Gwei"""
+    upper_limit_for_base_fee_in_gwei: float = float("inf")
+    """For type-2 transactions, raise an exception if `baseFee` is larger than this.
+    For legacy and type-1 transactions, raise an exception if `gasPrice` is
+    larger than this.  Default is no limit."""
+    gas_price_in_gwei: float = None
+    """Gas price, in gwei.  Relevant only for legacy and type-1 transactions.
+    Default is None, which means that the gas price will be estimated on chain."""
+    add_poa_support: bool = True
+    """Set to True to support to PoA chains like BnB, Polygon POS, Scroll, etc.  More details in class docstring."""
+    contract_address: str = None
+    """Address of smart contract"""
+    abi: dict[str, Any] = None
+    """ABI of smart contract; to read from a JSON file, use class method get_abi_json()"""
+    middlewares: List[Middleware] = []
+    """List of web3.py middlewares to use.  Will be added to the outermost layer of the onion, in the given order."""
+    rpc_logs: List[BaseRpcLog] = []
+    """Where to log RPC calls.  More details in class docstring"""
+
+    # Attributes that can only be set at instantiation
+    private_key: str = None
+    """Private key to use to send transactions"""
 
     # Derived attributes
     w3: Web3
@@ -110,45 +209,76 @@ class BaseClient:
     functions: ContractFunctions
     """ContractFunctions object of web3.py"""
 
-    # Class attributes
+    # Class-only attributes
     abi_dir: Union[Path, str] = Path(dirname(realpath(__file__))) / "abi"
     """Directory where to find the ABI json files"""
+    logger: Logger = getLogger("web3client.BaseClient")
+    """Class logger"""
+
+    # Constants
+    WEB3_BLOCK_VALIDATORS = deepcopy(BLOCK_VALIDATORS)
+    """Checks performed by web3.py on all blocks returned from the RPC.  This is used in web3.py's validation middleware."""
+    WEB3_BLOCK_FORMATTERS = deepcopy(BLOCK_FORMATTERS)
+    """Changes applied by web3.py to all blocks returned from the RPC.  Executed after WEB3_BLOCK_VALIDATORS."""
 
     def __init__(
         self,
-        node_uri: str,
+        node_uri: str = None,
         chain_id: int = None,
-        tx_type: int = 2,
+        tx_type: int = None,
         private_key: str = None,
-        max_priority_fee_in_gwei: float = 1,
-        upper_limit_for_base_fee_in_gwei: float = float("inf"),
+        max_priority_fee_in_gwei: float = None,
+        upper_limit_for_base_fee_in_gwei: float = None,
+        gas_price_in_gwei: float = None,
+        add_poa_support: bool = None,
         contract_address: str = None,
         abi: dict[str, Any] = None,
-        middlewares: List[Middleware] = [],
-        rpc_logs: List[BaseRpcLog] = [],
+        middlewares: List[Middleware] = None,
+        rpc_logs: List[BaseRpcLog] = None,
     ) -> None:
-        # Set the w3 client
-        self.set_provider(node_uri)
+        """
+        Create a new client.
 
-        # Set attributes.  The 'if' part is to allow subclasses to override
-        # the parameters.
-        if chain_id:
-            self.chain_id = chain_id
-        if tx_type:
-            self.tx_type = tx_type
+        The client can be instantiated with no arguments.  In this case, you will have
+        to call the following setters before using the client:
+
+         -  If `node_uri` is not given, call `self.set_provider()` before interacting
+                with the blockchain.
+         -  If `private_key` is not given, call `self.set_account()` before sending
+                transactions.
+         -  If `contract_address` is not given, call `self.set_contract()` before
+                using `self.call()` or `self.transact()`.
+        """
+        # Initialize the w3 client
+        self.set_provider(node_uri or self.__class__.node_uri)
+
+        # Initialize attributes without class-defined defaults
         if private_key:
             self.set_account(private_key)
-        if max_priority_fee_in_gwei:
-            self.max_priority_fee_in_gwei = max_priority_fee_in_gwei
-        if upper_limit_for_base_fee_in_gwei:
-            self.upper_limit_for_base_fee_in_gwei = upper_limit_for_base_fee_in_gwei
-        if abi:
-            self.abi = abi
-        if contract_address:
-            self.set_contract(contract_address)
-        if middlewares:
-            self.set_middlewares(middlewares)
-        if rpc_logs or self.__class__.rpc_logs:  # allow for class-defined rpc_logs
+
+        # Initialize attributes with class-defined defaults
+        self.chain_id = chain_id or self.__class__.chain_id
+        self.tx_type = tx_type or self.__class__.tx_type
+        self.max_priority_fee_in_gwei = (
+            max_priority_fee_in_gwei or self.__class__.max_priority_fee_in_gwei
+        )
+        self.upper_limit_for_base_fee_in_gwei = (
+            upper_limit_for_base_fee_in_gwei
+            or self.__class__.upper_limit_for_base_fee_in_gwei
+        )
+        self.gas_price_in_gwei = gas_price_in_gwei or self.__class__.gas_price_in_gwei
+        self.add_poa_support = (
+            add_poa_support
+            if add_poa_support is not None
+            else self.__class__.add_poa_support
+        )
+        self.maybe_add_poa_support()
+        self.abi = abi or self.__class__.abi
+        if contract_address or self.__class__.contract_address:
+            self.set_contract(contract_address or self.__class__.contract_address)
+        if middlewares or self.__class__.middlewares:
+            self.set_middlewares(middlewares or self.__class__.middlewares)
+        if rpc_logs or self.__class__.rpc_logs:
             self.set_rpc_logs(rpc_logs or self.__class__.rpc_logs)
 
         # Further initialization
@@ -165,12 +295,13 @@ class BaseClient:
 
     def set_provider(self, node_uri: str) -> Self:
         self.node_uri = node_uri
-        self.w3 = self.get_provider(node_uri)
+        self.w3 = self.make_provider(node_uri)
         return self
 
     def set_account(self, private_key: str) -> Self:
         """Make it possible for the client to sign transactions with the
         given private key"""
+        self.raise_if_provider_not_set()
         self.private_key = private_key
         self.account = Account.from_key(private_key)
         self.user_address = self.account.address
@@ -180,6 +311,7 @@ class BaseClient:
     def unset_account(self) -> Self:
         """Unset the previously added account, thus making it impossible
         for the client to sign transactions"""
+        self.raise_if_provider_not_set()
         self.private_key = None
         self.account = None
         self.user_address = None
@@ -187,6 +319,7 @@ class BaseClient:
         return self
 
     def set_contract(self, contract_address: str, abi: dict[str, Any] = None) -> Self:
+        self.raise_if_provider_not_set()
         abi = abi or self.abi
         if not abi:
             raise Web3ClientException("ABI not set")
@@ -196,12 +329,14 @@ class BaseClient:
         return self
 
     def set_middlewares(self, middlewares: List[Middleware]) -> Self:
+        self.raise_if_provider_not_set()
         self.middlewares = middlewares
-        for i, m in enumerate(middlewares):
-            self.w3.middleware_onion.inject(m, layer=i)
+        for m in middlewares:
+            self.w3.middleware_onion.inject(m, layer=0)
         return self
 
     def set_rpc_logs(self, rpc_logs: List[BaseRpcLog]) -> Self:
+        self.raise_if_provider_not_set()
         self.rpc_logs = rpc_logs
         for rpc_log in rpc_logs:
             self.w3.middleware_onion.add(construct_generic_rpc_log_middleware(rpc_log))
@@ -220,13 +355,14 @@ class BaseClient:
         """
         Build a basic transaction with type, nonce, chain ID and gas
 
-        - If not given, the nonce will be computed on chain
-        - If not given, the gas limit will be estimated on chain using gas_estimate()
+        - If not given, the nonce will be computed on chain using "eth_getTransactionCount"
+        - If not given, the gas limit will be estimated on chain using "eth_gasEstimate"
         - For type-2 transactions, if not given, the miner's tip (maxPriorityFeePerGas)
-          will be set to self.max_priority_fee_in_gwei
+           will be set to self.max_priority_fee_in_gwei
         - For type-2 transactions, the max gas fee is estimated according to the usual
-          formula maxMaxFeePerGas = 2 * baseFee + maxPriorityFeePerGas.
-        - For type-1 transactions, the gasPrice is estimated on-chain using eth_gasPrice.
+           formula maxMaxFeePerGas = 2 * baseFee + maxPriorityFeePerGas.
+        - For legacy and type-1 transactions, the gasPrice parameter, if not
+          provided at the instance level, will be estimated on chain using eth_gasPrice.
         """
 
         # Properties that you are not likely to change
@@ -235,18 +371,30 @@ class BaseClient:
             "from": self.user_address,
         }
 
+        # Infer tx_type if requested.  If the node does not support EIP-1559, fall back to legacy
+        tx_type = self.tx_type
+        if not tx_type:
+            tx_type = self.infer_tx_type()
+            self.logger.debug(f"Inferred tx_type={tx_type}")
+
         # Compute gas fee based on the transaction type
         gas_fee_in_gwei: float = None
 
-        # Pre EIP-1599, we only have gasPrice
-        if self.tx_type == 1:
-            self.w3.eth.set_gas_price_strategy(rpc.rpc_gas_price_strategy)
-            tx["gasPrice"] = self.w3.eth.generate_gas_price()
+        # For legacy transactions, we only have gasPrice
+        if tx_type == 0 or tx_type == 1:
+            if self.gas_price_in_gwei:
+                tx["gasPrice"] = Web3.to_wei(self.gas_price_in_gwei, "gwei")
+            else:
+                self.w3.eth.set_gas_price_strategy(rpc.rpc_gas_price_strategy)
+                tx["gasPrice"] = self.w3.eth.generate_gas_price()
             gas_fee_in_gwei = float(Web3.from_wei(tx["gasPrice"], "gwei"))
+            self.logger.debug(
+                f"Will build legacy TX with gasPrice={gas_fee_in_gwei} gwei"
+            )
 
         # Post EIP-1599, we have both the miner's tip and the max fee.
-        elif self.tx_type == 2:
-            tx["type"] = Web3.to_hex(self.tx_type)
+        elif tx_type == 2:
+            tx["type"] = Web3.to_hex(tx_type)
 
             # The miner tip is user-provided
             max_priority_fee_in_gwei = (
@@ -259,9 +407,12 @@ class BaseClient:
                 max_priority_fee_in_gwei
             )
             tx["maxFeePerGas"] = Web3.to_wei(maxFeePerGasInGwei, "gwei")
+            self.logger.debug(
+                f"Will build EIP-1559 TX with maxFeePerGas={maxFeePerGasInGwei} gwei maxPriorityFeePerGas={max_priority_fee_in_gwei} gwei"
+            )
         else:
             raise Web3ClientException(
-                f"Transaction with tx_type={self.tx_type} not supported, use either 1 or 2"
+                f"Transaction with tx_type={tx_type} not supported, use either 0, 1 or 2"
             )
 
         # Raise an exception if the fee is too high
@@ -736,8 +887,8 @@ class BaseClient:
 
     def estimate_gas_price_in_gwei(self) -> float:
         """
-        For Type-1 transactions (pre EIP-1559), estimate the gasPrice
-        parameter by asking it directly to the node.
+        For legacy and type-1 transactionss (pre EIP-1559), estimate the
+        gasPrice parameter by asking it directly to the node.
 
         Docs: https://web3py.readthedocs.io/en/stable/gas_price.html
         """
@@ -748,7 +899,7 @@ class BaseClient:
         """
         Raise an exception if the given gas fee in Gwei is too high.
 
-        For Type-1 transactions, pass the gasPrice; for Type-2,
+        For legacy and type-1 transactions, pass the gasPrice; for type-2,
         pass the baseFee.
         """
         if (
@@ -839,10 +990,13 @@ class BaseClient:
         Execute a contract function.
         This will write to the blockchain.
 
-        Example: transfer some tokens to the given address:
+        Example: transfer some ERC20 tokens to the given address:
+
+            ```
             client.transact(
                 client.functions.transfer(address, amount)
             )
+            ```
         """
         tx: TxParams = self.build_contract_tx(
             function, value_in_wei, nonce, gas_limit, max_priority_fee_in_gwei
@@ -872,12 +1026,42 @@ class BaseClient:
             private_key=self.private_key,
             max_priority_fee_in_gwei=self.max_priority_fee_in_gwei,
             upper_limit_for_base_fee_in_gwei=self.upper_limit_for_base_fee_in_gwei,
+            gas_price_in_gwei=self.gas_price_in_gwei,
             contract_address=self.contract_address,
             abi=self.abi,
             middlewares=self.middlewares,
             rpc_logs=self.rpc_logs,
         )
 
+    def supports_eip1559(self) -> bool:
+        """Return True if the chain supports EIP-1559"""
+        latest_block = self.w3.eth.get_block("latest")
+        return latest_block.get("baseFeePerGas") is not None
+
+    def infer_tx_type(self) -> int:
+        """Infer the transaction type for the chain: 2 if the chain supports
+        EIP-1559, else 0.  This is used to build transactions."""
+        if self.chain_id == 1:
+            return 2  # No need for ethereum mainnet ;-)
+        return 2 if self.supports_eip1559() else 0
+
+    def raise_if_provider_not_set(self) -> None:
+        """Raise an exception if the w3 provider is not set"""
+        if not self.w3:
+            raise ProviderNotSet(
+                "Web3 provider not set.  Did you pass a `node_uri` to web3client?"
+            )
+
+    def maybe_add_poa_support(self) -> None:
+        """If requested, enable support for PoA chains by disabling checks on
+        the `extraData` field of blocks."""
+        if self.add_poa_support:
+            BLOCK_VALIDATORS["extraData"] = lambda val: val
+            BLOCK_FORMATTERS["extraData"] = lambda val: HexBytes(val)
+        else:
+            BLOCK_VALIDATORS["extraData"] = self.WEB3_BLOCK_VALIDATORS["extraData"]
+            BLOCK_FORMATTERS["extraData"] = self.WEB3_BLOCK_FORMATTERS["extraData"]
+
     @staticmethod
     def get_contract(
         address: str,
@@ -918,7 +1102,7 @@ class BaseClient:
             return json.load(file)
 
     @staticmethod
-    def get_provider(node_uri: str) -> Web3:
+    def make_provider(node_uri: str) -> Web3:
         """
         Initialize provider (HTTPS & WS supported).
 
