@@ -25,7 +25,7 @@ supported_networks: List[NetworkConfig] = [
     # Avalanche C Chain
     {
         "name": "bnb",
-        "tx_type": 1,
+        "tx_type": 0,
         "chain_id": 56,
         "middlewares": [geth_poa_middleware],
         "rpcs": [
@@ -44,6 +44,15 @@ supported_networks: List[NetworkConfig] = [
         ],
         "coin": "AVAX",
     },
+    # Polygon
+    {
+        "name": "poly",
+        "tx_type": 2,
+        "chain_id": 137,
+        "middlewares": [geth_poa_middleware],
+        "rpcs": ["https://polygon-rpc.com"],
+        "coin": "MATIC",
+    },
     # Arbitrum One
     {
         "name": "arb",
@@ -63,11 +72,44 @@ supported_networks: List[NetworkConfig] = [
     # Kava EVM
     {
         "name": "kava",
-        "tx_type": 1,
+        "tx_type": 0,
         "chain_id": 2222,
         "rpcs": ["https://evm.kava.io"],
         "coin": "KAVA",
     },
+    # Optimism
+    {
+        "name": "opt",
+        "tx_type": 0,
+        "chain_id": 10,
+        "rpcs": ["https://mainnet.optimism.io"],
+        "coin": "ETH",
+    },
+    # Scroll
+    {
+        "name": "scroll",
+        "tx_type": 0,
+        "chain_id": 534352,
+        "middlewares": [geth_poa_middleware],
+        "rpcs": ["https://mainnet-rpc.scroll.io"],
+        "coin": "ETH",
+    },
+    # ZKFair
+    {
+        "name": "zkf",
+        "tx_type": 0,
+        "chain_id": 42766,
+        "rpcs": ["https://rpc.zkfair.io"],
+        "coin": "USDC",
+    },
+    # Manta Pacific
+    {
+        "name": "manta",
+        "tx_type": 0,
+        "chain_id": 169,
+        "rpcs": ["https://pacific-rpc.manta.network/http"],
+        "coin": "USDC",
+    },
 ]
 
 
